/* $Id: commands.c,v 1.1 2008/07/09 10:30:27 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/syslimits.h>
#include <sys/types.h>

#include <dirent.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "ukfs.h"
#include "extern.h"
#include "fsu_utils.h"

#define BUFSIZE (8192)

static int fsu_get_cwd(struct ukfs *, char *, int);

char cwd[PATH_MAX + 1], old_wd[PATH_MAX + 1];

command_t commands[] = {
	{ "cat",	fsu_cat_main	 },
	{ "cd",		fsu_cd_main	 },
	{ "chflags",	fsu_chflags_main },
	{ "chmod",	fsu_chmod_main	 },
	{ "chown",	fsu_chown_main	 },
	{ "cp",		fsu_cp_main	 },
	{ "diff",	fsu_diff_main	 },
	{ "dir",	fsu_ls_main	 },
	{ "du",		fsu_du_main	 },
	{ "exit",	fsu_exit_main	 },
	{ "help",	fsu_help_main	 },
	{ "ln",		fsu_ln_main	 },
	{ "ls",		fsu_ls_main	 },
	{ "mkdir",	fsu_mkdir_main	 },
	{ "mkfifo",	fsu_mkfifo_main	 },
	{ "mknod",	fsu_mknod_main	 },
	{ "mv",		fsu_mv_main	 },
	{ "pwd",	fsu_pwd_main	 },
	{ "rename",	fsu_rename_main	 },
	{ "rm",		fsu_rm_main	 },
	{ "rmdir",	fsu_rmdir_main	 },
	{ "touch",	fsu_touch_main	 },
	{ "umask",	fsu_umask_main	 },
	{ NULL,		NULL		 }
};

int
fsu_cd_main(struct ukfs *fs, int argc, char *argv[])
{
	int rv;
	char tmp[PATH_MAX + 1];
	
	if (argc == 1) {
		rv = strlcpy(old_wd, cwd, PATH_MAX + 1);
		if (rv == -1) {
			warn("%s: %s", argv[0], cwd);
			return -1;
		}

		cwd[0] = '/';
		cwd[1] = '\0';
	} else if (argc == 2 && strcmp(argv[1], "-") == 0) {
		rv = strlcpy(tmp, old_wd, PATH_MAX + 1);
		if (rv == -1) {
			warn("%s: %s", argv[0], old_wd);
			return -1;
		}

		strlcpy(old_wd, cwd, PATH_MAX + 1);
		if (rv == -1) {
			warn("%s: %s", argv[0], cwd);
			strlcpy(old_wd, tmp, PATH_MAX + 1);
			return -1;
		}

		rv = strlcpy(cwd, tmp, PATH_MAX + 1);
		if (rv == -1) {
			warn("%s: %s", argv[0], tmp);
			strlcpy(cwd, old_wd, PATH_MAX + 1);
			strlcpy(old_wd, tmp, PATH_MAX + 1);
			return -1;
		}
		printf("%s\n", cwd);
	} else if (argc == 2) {
		rv = ukfs_chdir(fs, argv[1]);
		if (rv == -1) {
			ukfs_chdir(fs, cwd);
			warn("%s: %s", argv[0], argv[1]);
			return -1;
		}
		strlcpy(old_wd, cwd, PATH_MAX + 1);

		fsu_get_cwd(fs, cwd, 0);
		
		return 0;
	} else {
		fprintf(stderr, "usage: cd path\n");
		return -1;
	}

	rv = ukfs_chdir(fs, cwd);
	if (rv == -1) {
		warn("%s: %s", argv[0], cwd);
		return -1;
	}
	return 0;
}

int
fsu_help_main(struct ukfs *fs, int argc, char *argv[])
{
	command_t *com;
	
	for (com = commands; com->c_name != NULL; com++)
		printf("%s\n", com->c_name);
	return 0;
}

int
fsu_exit_main(struct ukfs *fs, int argc, char *argv[])
{

	ukfs_release(fs, 1);
	exit(EXIT_SUCCESS);
}

int
fsu_umask_main(struct ukfs *fs, int argc, char **argv)
{
	mode_t mode, *set;

	switch (argc) {
	case 1:
		/*ukfs_umask((mode = ukfs_umask(0)));*/
		umask((mode = umask(0)));
		printf("%04o\n", mode);
		break;
	case 2:
		set = setmode(argv[1]);
		mode = getmode(set, ALLPERMS);
		free(set);
		/*ukfs_umask(mode);*/
		umask(mode);
		break;
	default:
		fprintf(stderr, "usage: %s [mode]\n", argv[0]);
		return -1;
	}
	return 0;
}

int
fsu_pwd_main(struct ukfs *fs, int argc, char *argv[])
{

	printf("%s\n", cwd);
	return 0;
}

#define ROOT_DIR (25)

static int
fsu_get_cwd(struct ukfs *fs, char *path, int level)
{
	int res, rv;
	off_t off;
	size_t len;
	struct dirent *dent;
	struct stat stdot, stdotdot;
	uint8_t buf[8192];
	bool found;

	rv = ukfs_stat(fs, "..", &stdotdot);
	if (rv == -1) {
		warn(".");
		return -1;
	}

	rv = ukfs_stat(fs, ".", &stdot);
	if (rv == -1) {
		warn(".");
		return -1;
	}

	if (stdot.st_ino == stdotdot.st_ino) {
		if (level == 0) {
			path[0] = '/';
			path[1] = '\0';
		}
		return ROOT_DIR;
	}

	dent = NULL;
	found = false;
	off = 0;
	while (!found) {
		rv = ukfs_getdents(fs, "..", &off, buf, sizeof(buf));
		if (rv == -1) {
			warn("..");
			return -1;
		} else if (rv == 0)
			break;
	
		dent = (struct dirent *)buf;
		while (!found && rv > 0) {
			if (dent->d_fileno == stdot.st_ino) {
				found = true;
				break;
			}
			
			rv -= _DIRENT_SIZE(dent);
			dent = _DIRENT_NEXT(dent);
		}
	}

	if (!found)
		return -1;

	rv = ukfs_chdir(fs, "..");
	if (rv == -1) {
		warn("..");
		return -1;
	}

	res = fsu_get_cwd(fs, path, level + 1);

	if (res != 0 && res != ROOT_DIR)
		return -1;

	if (res == ROOT_DIR) {
		path[0] = '/';
		rv = strlcpy(path+1, dent->d_name, PATH_MAX);
		if (rv != dent->d_namlen) {
			warn("%s", dent->d_name);
			return -1;
		}
	} else {
		len = strlen(path);
		len[path] = '/';
		len[path + 1] = '\0';
		rv = strlcat(path, dent->d_name, PATH_MAX + 1);
		if (rv != len + dent->d_namlen + 1) {
			warn("%s/%s", path, dent->d_name);
			return -1;
		}
	}

	if (level == 0) {
		rv = ukfs_chdir(fs, path);
		if (rv == -1) {
			warn("%s", path);
			return -1;
		}
	}
	return 0;
}
