/* $Id: fsu_chflags.c,v 1.1 2008/07/09 10:31:15 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <dirent.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <util.h>

#include "ukfs.h"

#include "fsu_flist.h"
#include "fsu_utils.h"
#include "fsu_chflags.h"

#define BUFSIZE (8192)

static int fsu_chflags_rec(struct ukfs *, const char *, u_long, u_long, int, 
			   int);

int
fsu_chflags_main(struct ukfs *fs, int argc, char *argv[])
{
	int cur_arg, flags, rv;
	char *comname;

	comname = argv[0];

	flags = fsu_chflags_parse_arg(&argc, &argv);
	if (flags == -1 || argc < 2) {
		fprintf(stderr, "usage: %s [-hRv] flag filename...\n", comname);
		return -1;
	}
	
	for (rv = 0, cur_arg = 1; cur_arg < argc; ++cur_arg)
		rv |= fsu_chflags(fs, argv[cur_arg], argv[0], flags);

	return rv;
}

int
fsu_chflags_parse_arg(int *argc, char ***argv)
{
	int flags, rv;

	flags = 0;
	while ((rv = getopt(*argc, *argv, "hRv")) != -1) {
		switch (rv) {
		case 'h':
			flags |= FSU_CHFLAGS_CHANGE_LINK;
			break;			
		case 'R':
			flags |= FSU_CHFLAGS_RECURSIVE;
			break;
		case 'v':
			flags |= FSU_CHFLAGS_VERBOSE;
			break;
		case '?':
		default: 
			return -1;
		}
	}
	*argc -= optind;
	*argv += optind;

	return flags;
}

int
fsu_chflags(struct ukfs *fs, const char *filename, const char *sflags, 
	    int flags)
{
	char *flagspp, *ep;
	int rv, oct;
	long val;
	u_long clear, set;
	struct stat filestat;

	rv = ukfs_lstat(fs, filename, &filestat);
	if (rv == -1) { 
		warn("lstat: %s", filename);
		return -1;
	}

	if (sflags[0] >= '0' && sflags[0] <= '7') {
		errno = 0;
		val = strtol(sflags, &ep, 8);
		if (val < 0)
			errno = ERANGE;
		if (errno) {
			warn("invalid flags: %s", sflags);
			return -1;
		}
		if (*ep) {
			warnx("invalid flags: %s", sflags);
			return -1;
		}
		set = val;
		oct = 1;
	} else {
		flagspp = strdup(sflags);
		rv = string_to_flags(&flagspp, &set, &clear);
		if (rv != 0) {
			warnx("invalid flag: %s", flagspp);
			return -1;
		}
		free(flagspp);
		clear = ~clear;
		oct = 0;
	}

	return fsu_chflags_ulong(fs, filename, set, clear, oct, flags);
}


int
fsu_chflags_ulong(struct ukfs *fs, const char *cpath, u_long set, u_long clear,
		  int oct, int flags)
{
	int res, rv;
	struct stat file_stat;
	u_long newflags;

	res = 0;

	if (flags & FSU_CHFLAGS_CHANGE_LINK)
		rv = ukfs_lstat(fs, cpath, &file_stat);
	else
		rv = ukfs_stat(fs, cpath, &file_stat);		
	
	if (rv == -1) {
		warn("stat: %s", cpath);
		return -1;
	}
	
	if (oct != 0)
		newflags = set;
	else
		newflags = (file_stat.st_mode | set) & clear;

	if (flags & FSU_CHFLAGS_CHANGE_LINK)
		res = ukfs_lchflags(fs, cpath, newflags);
	else
		res = ukfs_chflags(fs, cpath, newflags);

	if (flags & FSU_CHFLAGS_VERBOSE)
		printf("%s\n", cpath);
	
	if (!(flags & FSU_CHFLAGS_RECURSIVE) || !S_ISDIR(file_stat.st_mode))
		return res;
	
	res |= fsu_chflags_rec(fs, cpath, set, clear, oct, flags);

	return res;
}

static int
fsu_chflags_rec(struct ukfs *fs, const char *path, u_long set, u_long clear, 
		int oct, int flags)
{
	int flist_options, res, rv;
	FSU_FENT *root, *cur;
	u_long newflags;
	int (*chflags_fun)(struct ukfs *, const char *, u_long);

	if (flags & FSU_CHFLAGS_CHANGE_LINK) {
		chflags_fun = ukfs_lchflags;
		flist_options = FSU_FLIST_STATLINK;
	} else {
		chflags_fun = ukfs_chflags;
		flist_options = 0;
	}

	if (flags & FSU_CHFLAGS_RECURSIVE)
		flist_options |= FSU_FLIST_RECURSIVE;
	
	root = fsu_flist_build(fs, path, flist_options, NULL);
	if (root == NULL)
		return -1;
	
	for (res = rv = 0, cur = root->next; cur != NULL; cur = cur->next) {
		if (oct != 0)
			newflags = set;
		else
			newflags = (cur->sb.st_mode | set) & clear;

		rv = chflags_fun(fs, cur->path, newflags);
		if (rv == -1)
			warn("%s", cur->path);

		if (flags & FSU_CHFLAGS_VERBOSE)
			printf("%s\n", cur->path);
		res |= rv;
	}

	fsu_flist_free(root);

   	return res;
}
