/* $Id: fsu_chown.c,v 1.1 2008/07/09 10:32:01 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <dirent.h>

#include <err.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include "ukfs.h"

#include "fsu_flist.h"
#include "fsu_utils.h"
#include "fsu_chown.h"

#define BUFSIZE (8192)

static int fsu_chown_rec(struct ukfs *, const char *path, uid_t, gid_t, int);

int
fsu_chown_main(struct ukfs *fs, int argc, char *argv[])
{
	int cur_arg, flags, rv;
	char *comname;

	comname = argv[0];

	flags = fsu_chown_parse_arg(&argc, &argv);
	if (flags == -1 || argc < 2) {
		fprintf(stderr, 
			"usage: %s [-hRv] [owner][:group] filename...\n", 
			comname);
		return -1;
	}
	
	for (rv = 0, cur_arg = 1; cur_arg < argc; ++cur_arg)
		rv |= fsu_chown(fs, argv[cur_arg], argv[0], flags);

	return rv;
}

int
fsu_chown_parse_arg(int *argc, char ***argv)
{
	int flags, rv;

	flags = 0;
	while ((rv = getopt(*argc, *argv, "hRv")) != -1) {
		switch (rv) {
		case 'h':
			flags |= FSU_CHOWN_CHANGE_LINK;
			break;
		case 'R':
			flags |= FSU_CHOWN_RECURSIVE;
			break;
		case 'v':
			flags |= FSU_CHOWN_VERBOSE;
			break;
		case '?':
		default: 
			return -1;
		}
	}
	*argc -= optind;
	*argv += optind;

	return flags;
}

int
fsu_chown(struct ukfs *fs, const char *filename, const char *ids, int flags)
{
	uid_t uid;
	gid_t gid;
	char *grp;

	uid = gid = -1;

	grp = strchr(ids, ':');
	if (grp != NULL) {
		*grp++ = '\0';
		gid_from_group(grp, &gid);
	} else {
		grp = strrchr(ids, '.');
		if (grp != NULL) {
			*grp++ = '\0';
			gid_from_group(grp, &gid);
		}
	}
	uid_from_user(ids, &uid);

	return fsu_chown_ids(fs, filename, uid, gid, flags);
}


int
fsu_chown_ids(struct ukfs *fs, const char *cpath, uid_t uid, gid_t gid, 
	      int flags)
{
	int rv;
	struct stat file_stat;
	
	rv = ukfs_lstat(fs, cpath, &file_stat);
	if (rv == -1) {
		warn("%s", cpath);
		return -1;
	}
		
	if (flags & FSU_CHOWN_CHANGE_LINK)
		rv = ukfs_lchown(fs, cpath, uid, gid);
	else
		rv = ukfs_chown(fs, cpath, uid, gid);
		
	if (rv == -1) {
		warn("%s", cpath);
		return -1;
	}
	
	if (flags & FSU_CHOWN_VERBOSE)
		printf("%s\n", cpath);
	
	if (!(flags & FSU_CHOWN_RECURSIVE) || !S_ISDIR(file_stat.st_mode))
		return 0;
	

	return fsu_chown_rec(fs, cpath, uid, gid, flags);
}

static int
fsu_chown_rec(struct ukfs *fs, const char *path, uid_t uid, gid_t gid, 
	      int flags)
{
	int flist_options, res, rv;
	FSU_FENT *root, *cur;
	int (*chown_fun)(struct ukfs *, const char *, uid_t, gid_t);


	if (flags & FSU_CHOWN_CHANGE_LINK) {
		chown_fun = ukfs_lchown;
		flist_options = FSU_FLIST_STATLINK;
	} else {
		chown_fun = ukfs_chown;
		flist_options = 0;
	}

	if (flags & FSU_CHOWN_RECURSIVE)
		flist_options |= FSU_FLIST_RECURSIVE;

	root = fsu_flist_build(fs, path, flist_options, NULL);
	if (root == NULL)
		return -1;

	for (res = 0, cur = root; cur != NULL; cur = cur->next) {
		if (flags & FSU_CHOWN_VERBOSE)
			printf("%s\n", cur->path);

		rv = chown_fun(fs, cur->path, uid, gid);

		if (rv == -1)
			warn("%s", cur->path);
		res |= rv;
	}

	fsu_flist_free(root);

	return res;
}
