/* $Id: fsu_diff.c,v 1.1 2008/07/09 10:32:24 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/syslimits.h>

#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ukfs.h"

#include "fsu_flist.h"
#include "fsu_utils.h"
#include "fsu_mount.h"
#include "fsu_diff.h"

#define BUFSIZE (8192)

static int fsu_diff_file(struct ukfs *, const FSU_FENT *, const FSU_FENT *);
static int fsu_diff_fent(struct ukfs *, const FSU_FENT *, const FSU_FENT *);
static int fsu_diff_link(struct ukfs *, const FSU_FENT *, const FSU_FENT *);
int	   fsu_diff_namecmp(const FSU_FENT **, const FSU_FENT **);

int
fsu_diff_main(struct ukfs *fs, int argc, char *argv[])
{
	int flags, rv;

	flags = fsu_diff_parse_arg(&argc, &argv);

	if (argc != 2) {
		fprintf(stderr, "usage: %s filename1 filename2\n", argv[0]);
		return -1;
	}

	rv = fsu_diff(fs, argv[0], argv[0], flags);

	return rv;
}

int
fsu_diff_parse_arg(int *argc, char ***argv)
{
	int flags, ch;

	flags = 0;
	while ((ch = getopt(*argc, *argv, "Rr")) != -1) {
		switch (ch) {
		case 'r':
		case 'R':
			flags |= FSU_DIFF_RECURSIVE;
			break;
		case '?':
		default:
			return -1;
		}
	}

	*argc -= optind;
	*argv += optind;

	return flags;
}

int
fsu_diff(struct ukfs *fs, const char *p1, const char *p2, int flags)
{
	FSU_FENT *r1, *c1, *r2, *c2;
	int flist_options, rv;

	flist_options = FSU_FLIST_STATLINK;
	if (flags & FSU_DIFF_RECURSIVE)
		flist_options |= FSU_FLIST_RECURSIVE;

	r1 = fsu_flist_build(fs, p1, flist_options, fsu_diff_namecmp);
	if (r1 == NULL)
		return -1;

	r2 = fsu_flist_build(fs, p2, flist_options, fsu_diff_namecmp);
	if (r2 == NULL) {
		fsu_flist_free(r1);
		return -1;
	}

	if (r1->next == NULL && r2->next == NULL) {
		c1 = r1;
		c2 = r2;
	} else if (r1->next == NULL || r2->next == NULL) {
		rv = -1;
		goto out;
	} else {
		c1 = r1->next;
		c2 = r2->next;
	}

	for (; c1 != NULL && c2 != NULL; c1 = c1->next, c2 = c2->next) {
		rv = fsu_diff_fent(fs, c1, c2);
		if (rv != 0) {
			fprintf(stderr, "%s and %s differ\n", 
				c1->path, c2->path);
			rv = -1;
			goto out;
		}
		
	}
	if (c1 != NULL) {
		fprintf(stderr, "%s only on %s\n", c1->path, p1);
		rv = -1;
		goto out;
	} else if (c2 != NULL) {
		fprintf(stderr, "%s only on %s\n", c2->path, p2);
		rv = -1;
		goto out;
	}
	rv = 0;
out:
	fsu_flist_free(r1);
	fsu_flist_free(r2);
	return rv;
}

static int 
fsu_diff_fent(struct ukfs *fs, const FSU_FENT *f1, const FSU_FENT *f2)
{
	if ((f1->sb.st_mode & S_IFMT) != (f2->sb.st_mode & S_IFMT))
		return -1;

	switch (f1->sb.st_mode & S_IFMT) {
	case S_IFIFO:
		return 0;
	case S_IFCHR:
	case S_IFBLK:
		return (f1->sb.st_rdev != f2->sb.st_rdev);
	case S_IFDIR:
		if (f1->childno != f2->childno)
			return -1;
		return strcmp(f1->filename, f2->filename);
	case S_IFREG:
		return fsu_diff_file(fs, f1, f2);
	case S_IFLNK:
		return fsu_diff_link(fs, f1, f2);
	default:
		return -1;
	}

}

static int
fsu_diff_file(struct ukfs *fs, const FSU_FENT *f1, const FSU_FENT *f2)
{
	uint8_t buf1[BUFSIZE], buf2[BUFSIZE];
	int rd1, rd2;
	off_t off, diff;

	if (f1->sb.st_size != f2->sb.st_size)
		return -1;
	if (f1->parent != NULL && strcmp(f1->filename, f2->filename) != 0)
		return -1;

	off = 0;
	do {
		rd1 = ukfs_read(fs, f1->path, off, buf1, BUFSIZE);
		if (rd1 == -1) {
			warn("%s", f1->path);
			return -1;
		}

		rd2 = ukfs_read(fs, f2->path, off, buf2, BUFSIZE);
		if (rd2 == -1) {
			warn("%s", f2->path);
			return -1;
		}

		if (rd1 != rd2)
			return 1;
		
		diff = memcmp(buf1, buf2, rd1);
		if (diff != 0)
			return 1;
		
		off += rd1;
	} while (rd1 > 0);
	return 0;
}

static int
fsu_diff_link(struct ukfs *fs, const FSU_FENT *f1, const FSU_FENT *f2)
{
	char buf1[PATH_MAX + 1], buf2[PATH_MAX + 1];
	int rd1, rd2;

	if (strcmp(f1->filename, f2->filename) != 0)
		return -1;

	rd1 = ukfs_readlink(fs, f1->path, buf1, PATH_MAX);
	if (rd1 == -1) {
		warn("%s", f1->path);
		return -1;
	}
	rd2 = ukfs_readlink(fs, f2->path, buf2, PATH_MAX);
	if (rd2 == -1) {
		warn("%s", f2->path);
		return -1;
	}		
	if (rd1 != rd2 || strncmp(buf1, buf2, rd1) != 0)
		return -1;

	return 0;
}

int
fsu_diff_namecmp(const FSU_FENT **a, const FSU_FENT **b)
{

	return (strcmp((*a)->filename, (*b)->filename));
}
