/* $Id: fsu_du.c,v 1.1 2008/07/09 10:32:40 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <dirent.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ukfs.h"

#include "fsu_flist.h"
#include "fsu_utils.h"
#include "fsu_du.h"

#define BUFSIZE (8192)


static void print_info(FSU_FENT *, int);

int
fsu_du_main(struct ukfs *fs, int argc, char *argv[])
{
	int cur_arg, flags, rv;
	char *comname;

	comname = argv[0];

	flags = fsu_du_parse_arg(&argc, &argv);
	if (flags == -1 || argc < 0) {
		fprintf(stderr, "usage: %s [-aghkms] [filename...]\n", 
			comname);
		return -1;
	}

	if (argc == 0)
		return fsu_du(fs, ".", flags);
		
	for (rv = 0, cur_arg = 0; cur_arg < argc; ++cur_arg)
		rv |= fsu_du(fs, argv[cur_arg], flags);

	return rv;
}

int
fsu_du_parse_arg(int *argc, char ***argv)
{
	int flags, rv;

	flags = 0;
	while ((rv = getopt(*argc, *argv, "aghkms")) != -1) {
		switch (rv) {
		case 'a':
			flags |= FSU_DU_AFLAG;
			break;
		case 'g':
			flags |= FSU_DU_GFLAG;
			flags &= ~(FSU_DU_HFLAG | FSU_DU_KFLAG | FSU_DU_MFLAG);
			break;
		case 'h':
			flags |= FSU_DU_HFLAG;
			flags &= ~(FSU_DU_GFLAG | FSU_DU_KFLAG | FSU_DU_MFLAG);
			break;
		case 'k':
			flags |= FSU_DU_KFLAG;
			flags &= ~(FSU_DU_GFLAG | FSU_DU_HFLAG | FSU_DU_MFLAG);
			break;
		case 'm':
			flags |= FSU_DU_MFLAG;
			flags &= ~(FSU_DU_GFLAG | FSU_DU_HFLAG | FSU_DU_KFLAG);
			break;
		case 's':
			flags |= FSU_DU_SFLAG;
			break;
		case '?':
		default:
			return -1;
		}
	}
	*argc -= optind;
	*argv += optind;

	return flags;
}

int
fsu_du(struct ukfs *fs, const char *path, int flags)
{
	int flist_options, res;
	FSU_FENT *root, *cur;

	res = 0;

	flist_options = FSU_FLIST_RECURSIVE | FSU_FLIST_STATLINK | 
		FSU_FLIST_POSTDIR | FSU_FLIST_SHOWDOT;

	root = fsu_flist_build(fs, path, flist_options, NULL);
	if (root == NULL)
		return -1;

	for (cur = root->next; cur != NULL; cur = cur->next) {
		if (cur->parent == NULL) {
			
		} else if (S_ISDIR(cur->sb.st_mode)) {
			cur->parent->size += cur->size + cur->sb.st_blocks;
		} else {
			cur->size = cur->sb.st_blocks;
			cur->parent->size += cur->size;
		}
		if (cur->isdot)
			continue;
		if (cur->parent == NULL || 
		    flags & FSU_DU_AFLAG || 
		    (S_ISDIR(cur->sb.st_mode) && !(flags & FSU_DU_SFLAG)))
			print_info(cur, flags);
	}

	fsu_flist_free(root);
	return res;
}

static void
print_info(FSU_FENT *file, int flags)
{
	char buf[5];
	unsigned long bsize;
	int64_t size;


	if (flags & FSU_DU_HFLAG) {
		size = file->size * 512;
		humanize_number(buf, sizeof(buf), 
				size, "", 
				HN_AUTOSCALE,
				HN_B | HN_NOSPACE | HN_DECIMAL);
		
		printf("%s\t%s\n", buf, file->path);
		return;
	}

	if (flags & FSU_DU_KFLAG)
		bsize = 2;
	else if (flags & FSU_DU_MFLAG)
		bsize = 2 * 1024;
	else if (flags & FSU_DU_GFLAG)
		bsize = 2 * 1024 * 1024;
	else
		bsize = 1;

	printf("%lld\t%s\n",
	       (long long)howmany(file->size, (int64_t)bsize),
	       file->path);
}
