/* $Id: fsu_ls.c,v 1.1 2008/07/09 10:33:19 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <dirent.h>

#include <sys/syslimits.h>

#include <err.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <tzfile.h>
#include <unistd.h>

#include "ukfs.h"

#include "fsu_utils.h"
#include "fsu_ls.h"
#include "fsu_flist.h"

#define BUFSIZE (8192)

static int fsu_ls_rec(struct ukfs *, const char *, int);
static void print_dent(struct ukfs *, FSU_FENT *, int);
static void print_long(struct ukfs *, FSU_FENT *, int);
static void print_type(mode_t, int);
static void print_time(time_t, int);
static void print_long_file(struct ukfs *, const char *, struct stat *, int);

int
fsu_ls_main(struct ukfs *fs, int argc, char *argv[])
{
	int cur_arg, flags = 0, rv;
	char *comname;

	comname = argv[0];
	flags = fsu_ls_parse_arg(&argc, &argv);

	if (flags == -1) {
		fprintf(stderr, "usage: "
			"%s [-AacdFlnpRrSTtu] [dirname...]\n",
			comname);
		return -1;
	} 

	if (argc == 0)
		return fsu_ls(fs, ".", flags);

	for (rv = 0, cur_arg = 0; cur_arg < argc; ++cur_arg) {
		rv |= fsu_ls(fs, argv[cur_arg], flags);
		if (cur_arg != argc - 1)
			printf("\n");
	}
	return rv;
}

int
fsu_ls_parse_arg(int *argc, char ***argv)
{
	int flags, rv;

	flags = 0;
	if (getuid() == 0)
		flags |= FSU_LS_SHOW_ALL_BUT_DOT;

	while ((rv = getopt(*argc, *argv, "AacdFlnpRrSTtu")) != -1) {
		switch (rv) {
		case 'A':
			flags |= FSU_LS_SHOW_ALL_BUT_DOT;
			break;
		case 'a':
			flags |= FSU_LS_SHOW_ALL;
			break;
		case 'c':
			flags |= FSU_LS_CHANGED_TIME;
			flags &= ~(FSU_LS_ACCESS_TIME | FSU_LS_SORT_SIZE | 
				   FSU_LS_SORT_MODIFIED);
			break;
		case 'd':
			flags |= FSU_LS_DIR_AS_FILE;
			break;
		case 'F':
			flags |= FSU_LS_DISPLAY_TYPE;
			break;
		case 'l':
			flags |= FSU_LS_LONG;
			break;
		case 'n':
			flags |= FSU_LS_NUMID;
			break;
		case 'p':
			flags |= FSU_LS_DISPLAY_DIRTYPE;
			break;
		case 'R':
			flags |= FSU_LS_RECURSIVE;
			break;
		case 'r':
			flags |= FSU_LS_REVERSE;
			break;
		case 'S':
			flags |= FSU_LS_SORT_SIZE;
			flags &= ~(FSU_LS_ACCESS_TIME | FSU_LS_CHANGED_TIME |
				   FSU_LS_SORT_MODIFIED);
			break;
		case 'T':
			flags |= FSU_LS_FULLTIME;
			break;
		case 't':
			flags |= FSU_LS_SORT_MODIFIED;
			flags &= ~(FSU_LS_ACCESS_TIME | FSU_LS_CHANGED_TIME | 
				   FSU_LS_SORT_SIZE);
			break;
		case 'u':
			flags |= FSU_LS_ACCESS_TIME;
			flags &= ~(FSU_LS_CHANGED_TIME | FSU_LS_SORT_SIZE | 
				   FSU_LS_SORT_MODIFIED);
			break;
		case '?':
		default:
			return -1;
		}
	}
	*argv += optind;
	*argc -= optind;

	return flags;
}

int
fsu_ls(struct ukfs *fs, const char *cpath, int flags)
{
	int rv;
	struct stat file_stat;

	rv = ukfs_lstat(fs, cpath, &file_stat);
	if (rv == -1) {
		warn("%s", cpath);
		return -1;
	}
	
	if (flags & FSU_LS_DIR_AS_FILE || !S_ISDIR(file_stat.st_mode)) {
		if (flags & FSU_LS_LONG)
			print_long_file(fs, cpath, &file_stat, flags);
		else
			printf("%s\n", cpath);
		return 0;
	}
	
	return fsu_ls_rec(fs, cpath, flags);
}

static int
fsu_ls_rec(struct ukfs *fs, const char *path, int flags)
{
	FSU_FENT *root, *cur, *curdir;
	int (*cmpfun)(const FSU_FENT **, const FSU_FENT **);
	int flist_options, sort;
	struct stat sb;

	sort = flags & (FSU_LS_SORT_SIZE  | FSU_LS_ACCESS_TIME | 
			FSU_LS_CHANGED_TIME | FSU_LS_SORT_MODIFIED);

	if (flags & FSU_LS_REVERSE) {
		switch (sort) {
		case FSU_LS_SORT_SIZE:
			cmpfun = revsizecmp;
			break;
		case FSU_LS_ACCESS_TIME:
			cmpfun = revacccmp;
			break;
		case FSU_LS_CHANGED_TIME: 
			cmpfun = revstatcmp;
			break;
		case FSU_LS_SORT_MODIFIED:
			cmpfun = revmodcmp;
			break;
		default:
			cmpfun = revnamecmp;
		}
	} else {
		switch (sort) {
		case FSU_LS_SORT_SIZE:
			cmpfun = sizecmp;
			break;
		case FSU_LS_ACCESS_TIME:
			cmpfun = acccmp;
			break;
		case FSU_LS_CHANGED_TIME: 
			cmpfun = statcmp;
			break;
		case FSU_LS_SORT_MODIFIED:
			cmpfun = modcmp;
			break;
		default:
			cmpfun = namecmp;
		}	
	}
	
	sort = ukfs_stat(fs, ".", &sb);
	if (sort == -1) {
		warn(".");
		return -1;
	}

	flist_options = FSU_FLIST_STATLINK;
	if (flags & FSU_LS_RECURSIVE)
		flist_options |= FSU_FLIST_RECURSIVE;
	if (flags & FSU_LS_SHOW_ALL)
		flist_options |= FSU_FLIST_SHOWDOT;

	root = fsu_flist_build(fs, path,flist_options, cmpfun);
	if (root == NULL)
		return -1;

	if (flags & FSU_LS_SHOW_ALL_BUT_DOT)
		flags |= FSU_LS_SHOW_ALL;
	
	curdir = root;
	while (curdir != NULL) {
		if (curdir->sb.st_ino != sb.st_ino)
			printf("%s:\n", curdir->path);
		for (cur = curdir->next; cur != NULL; cur = cur->next)
			if (cur->parent == curdir && 
			    (cur->filename[0] != '.' || 
			     flags & FSU_LS_SHOW_ALL))
				print_dent(fs, cur, flags);
	
		if (!(flags & FSU_LS_RECURSIVE))
			break;

		while ((curdir = curdir->next) != NULL && 
		       (!S_ISDIR(curdir->sb.st_mode) || curdir->isdot))
			continue;

		if (curdir != NULL)
			printf("\n");
	}

	fsu_flist_free(root);
	
	return 0;
}

static void
print_dent(struct ukfs *fs, FSU_FENT *file, int flags)
{
	if (flags & FSU_LS_LONG)
		print_long(fs, file, flags);
	else
		printf("%s", file->filename);
	if (!(flags & FSU_LS_LONG && S_ISLNK(file->sb.st_mode)))
		print_type(file->sb.st_mode, flags);
	printf("\n");
}

static void
print_type(mode_t mode, int flags)
{
	if ((flags & FSU_LS_DISPLAY_DIRTYPE) && S_ISDIR(mode))
		printf("/");
	
	if (!(flags & FSU_LS_DISPLAY_TYPE))
		return;

	switch (mode & S_IFMT) {
	case S_IFDIR:
		printf("/");
		return;
	case S_IFIFO:
		printf("|");
		return;
	case S_IFLNK:
		printf("@");
		return;
	case S_IFSOCK:
		printf("=");
		return;
	case S_IFWHT:
		printf("%%");
		return;
	}
	if (mode & (S_IXUSR | S_IXGRP | S_IXOTH))
		printf("*");
}
static void
print_long(struct ukfs *fs, FSU_FENT *file, int flags)
{
	char mode[12];
	time_t ftime;
     
	strmode(file->sb.st_mode, mode);

	printf("%s %i\t", mode, file->sb.st_nlink);

	if (flags & FSU_LS_NUMID)
		printf("%i %i\t", file->sb.st_uid, file->sb.st_gid);
	else
		printf("%s %s\t", user_from_uid(file->sb.st_uid, 0),
		       group_from_gid(file->sb.st_gid, 0));

	if (S_ISCHR(file->sb.st_mode) || S_ISBLK(file->sb.st_mode))
		printf("%i, %i\t", (file->sb.st_rdev & 0x000fff00)>>8,
		       ((file->sb.st_rdev & 0xfff00000) >> 12) |
		       (file->sb.st_rdev & 0x000000ff) >>  0);
	else
		printf("%i\t", (int)file->sb.st_size);

	switch (flags & (FSU_LS_CHANGED_TIME | FSU_LS_ACCESS_TIME)) {
	case FSU_LS_CHANGED_TIME:
		ftime = file->sb.st_ctime;
		break;
	case FSU_LS_ACCESS_TIME:
		ftime = file->sb.st_atime;
		break;
	default:
		ftime =	file->sb.st_mtime;
	}

	print_time(ftime, flags);
	
	printf("%s", file->filename);
	if (S_ISLNK(file->sb.st_mode)) {
		char lnkname[PATH_MAX+1];
		int rv;

		rv = ukfs_readlink(fs, file->path, lnkname, PATH_MAX);
		if (rv == -1) {
			warn("%s", file->path);
			return;
		}
		lnkname[rv] = '\0';
		printf(" -> %s", lnkname);
	}
}

static void
print_long_file(struct ukfs *fs, const char *path, struct stat *filestat, 
		int flags)
{
	FSU_FENT file;
	char *p;

	file.path = strdup(path);

	p = strrchr(file.path, '/');
	if (p == NULL || (p == file.path && file.path[1] == '\0'))
		file.filename = file.path;
	else
		file.filename = p + 1;

	memcpy(&(file.sb), filestat, sizeof(struct stat));
	print_long(fs, &file, flags);
	free(file.path);
	printf("\n");
}

static void
print_time(time_t ftime, int flags)
{
	char timestring[26];
	time_t now;

	now = time(NULL);
	ctime_r(&ftime, timestring);

	/* remove '\n' */
	timestring[24] = '\0';

	if (flags & FSU_LS_FULLTIME) {
		printf("%s ", timestring);
		return;
	}

#define	SIXMONTHS ((DAYSPERNYEAR / 2) * SECSPERDAY)

	if (ftime + SIXMONTHS > now && ftime - SIXMONTHS < now) {
		timestring[16] = '\0';
		printf("%s ", timestring+4);
	} else {
		timestring[11] = '\0';
		printf("%s %s ", timestring+4, timestring+20);
	}
}
