/* $Id: fsu_rm.c,v 1.1 2008/07/09 10:34:33 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>

#include <assert.h>
#include <err.h>
#include <stdio.h>
#include <unistd.h>

#include "ukfs.h"

#include "fsu_flist.h"
#include "fsu_utils.h"
#include "fsu_rm.h"

#define BUFSIZE (8192)

static int fsu_rm_rec(struct ukfs *, const char *, int);

int
fsu_rm_main(struct ukfs *fs, int argc, char *argv[])
{
	int cur_arg, flags, rv;
	char *comname;

	comname = argv[0];

	flags = fsu_rm_parse_arg(&argc, &argv);
	if (flags == -1 || argc < 1) {
		fprintf(stderr, "usage: %s [-Rrv] file...\n", comname);
		return -1;
	}

	for (rv = 0, cur_arg = 0; cur_arg < argc; ++cur_arg)
		rv |= fsu_rm(fs, argv[cur_arg], flags);

	return rv;
}

int
fsu_rm_parse_arg(int *argc, char ***argv)
{
	int flags, rv;

	flags = 0;
	while ((rv = getopt(*argc, *argv, "Rrv")) != -1) {
		switch (rv) {
		case 'R':
		case 'r': 
			flags |= FSU_RM_RECURSIVE;
			break;
		case 'v':
			flags |= FSU_RM_VERBOSE;
			break;
		case '?':
		default: 
			return -1;
		}
	}
	*argc -= optind;
	*argv += optind;

	return flags;
}

int
fsu_rm(struct ukfs *fs, const char *cpath, int flags)
{
	int rv;
	struct stat file_stat;

	rv = ukfs_lstat(fs, cpath, &file_stat);
	if (rv == -1) {
		warn("%s", cpath);
		return -1;
	}
	
	if (!S_ISDIR(file_stat.st_mode)) {
		rv = ukfs_remove(fs, cpath);
		if (rv == -1) {
			warn("%s", cpath);
			return -1;
		}
		return 0;
	} else if (!(flags & FSU_RM_RECURSIVE)) {
		warnx("%s: is a directory", cpath);
		return -1;
	}
	
	return fsu_rm_rec(fs, cpath, flags);
}

static int
fsu_rm_rec(struct ukfs *fs, const char *path, int flags)
{
	int flist_options, res, rv;
	FSU_FENT *root, *cur;
	

	flist_options = FSU_FLIST_STATLINK | FSU_FLIST_POSTDIR;
	if (flags & FSU_RM_RECURSIVE)
		flist_options |= FSU_FLIST_RECURSIVE;

	root = fsu_flist_build(fs, path,flist_options, NULL);
	if (root == NULL)
		return -1;

	for (res = 0, cur = root; cur != NULL; cur = cur->next) {
		if (flags & FSU_RM_VERBOSE)
			printf("%s\n", cur->path);

		if (S_ISDIR(cur->sb.st_mode))
			rv = ukfs_rmdir(fs, cur->path);
		else
			rv = ukfs_remove(fs, cur->path);
	
		if (rv == -1) {
			warn("%s", cur->path);
			rv = -1;
			break;
		}
		res |= rv;
	}

	fsu_flist_free(root);

	return res;
}
