/* $Id: fsu_flist.c,v 1.1 2008/07/09 10:35:22 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <dirent.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/syslimits.h>

#include "ukfs.h"

#include "fsu_dirent.h"
#include "fsu_flist.h"

#define ISDOT(a) ((a)[0] == '.' && \
		  ((a)[1] == '\0' || ((a)[1] == '.' && (a)[2] == '\0')))

static void fsu_flist_sort(FSU_FENT **, FSU_FENT **, int);
static FSU_FENT *fsu_flist_alloc(struct ukfs *, struct dirent *, FSU_FENT *);
static FSU_FENT *fsu_flist_alloc_root(struct ukfs *, const char *);

static FSU_FENT **array;
static unsigned int array_size;
static int (*cmp)(const void *, const void *);
static int (*statfun)(struct ukfs *, const char *, struct stat *);

FSU_FENT
*fsu_flist_build(struct ukfs *fs, const char *rootp, int flags, 
		 int (*f)(const FSU_FENT **, const FSU_FENT **))
{
	FSU_FENT *cur, *first, *last, *child, *root, *prev;
	FSU_DIR *curdir;
	struct dirent *dent;

	if (fs == NULL || rootp == NULL)
		return NULL;

	statfun = flags & FSU_FLIST_STATLINK ? ukfs_lstat : ukfs_stat;
	cmp = (int (*)(const void *, const void *))f;
	array_size = 0;
	array = NULL;

	root = fsu_flist_alloc_root(fs, rootp);
	if (root == NULL)
		return NULL;

	for (prev = cur = root; cur != NULL;) {
		if (cur->isdot || !S_ISDIR(cur->sb.st_mode) || cur->done)
			goto next;

		if (cur == prev && cur == root->next)
			prev = root;

		curdir = fsu_opendir(fs, cur->path);
		if (curdir == NULL)
			goto next;

		first = last = NULL;			
		while ((dent = fsu_readdir(curdir)) != NULL) {
			if (!(flags & FSU_FLIST_SHOWDOT) && ISDOT(dent->d_name))
				continue;
			
			child = fsu_flist_alloc(fs, dent, cur);
			if (child == NULL)
				continue;
			
			if(first == NULL)
				first = last = child;
			else
				last = last->next = child;
			
			cur->childno++;	
		}
		
		fsu_closedir(curdir);
		cur->done = true;

		if(last != NULL) {
			if (last != first && cmp != NULL)
				fsu_flist_sort(&first, &last, cur->childno);
			if (flags & FSU_FLIST_POSTDIR) {
				if (cur == root) {
					last->next = root;
					prev = root = first;
				} else {
					last->next = cur;
					prev->next = first;
				}
				cur = first;
				if (!(flags & FSU_FLIST_RECURSIVE))
					break;

				continue;
			} else {
				last->next = cur->next;
				cur->next = first;
			}
		}

		if (!(flags & FSU_FLIST_RECURSIVE))
			break;
	next:
		prev = cur;
		cur = cur->next;
	}

	if (array != NULL) {
		free(array);
		array_size = 0;
	}

	return root;
}

static FSU_FENT
*fsu_flist_alloc(struct ukfs *fs, struct dirent *dent, FSU_FENT *parent)
{
	FSU_FENT *child;
	int rv;
	bool is_child_of_slash;


	is_child_of_slash = (parent->path[0] == '/' && parent->path[1] == '\0');

	child = malloc(sizeof(FSU_FENT));
	if (child == NULL) {
		warn("malloc");
		return NULL;
	}
	
	child->isdot = ISDOT(dent->d_name);
	child->parent = parent;
	child->size = 0;
	child->done = false;
	if (is_child_of_slash)
		child->pathlen = parent->pathlen + dent->d_namlen;
	else
		child->pathlen = parent->pathlen + dent->d_namlen + 1;

	child->path = malloc(child->pathlen + 1);
	if (child->path == NULL) {
		free(child);
		return NULL;
	}

	if (is_child_of_slash)
		rv = snprintf(child->path, child->pathlen + 1, "/%s", 
			      dent->d_name);
	else
		rv = snprintf(child->path, child->pathlen + 1, "%s/%s", 
			      parent->path, dent->d_name);
		
	if (rv != child->pathlen) {
		warn("%s/%s", is_child_of_slash ? "" : parent->path, 
		     dent->d_name);
		fsu_flist_free_entry(child);
		return NULL;
	}

	child->filename = child->path + child->pathlen - dent->d_namlen;
	child->next = NULL;
	child->childno = 0;
	child->level = child->parent->level + 1;
	
	rv = statfun(fs, child->path, &child->sb);
	if (rv == -1) {
		warn("%s", child->path);
		fsu_flist_free_entry(child);
		return NULL;
	}

	return child;
}

static FSU_FENT
*fsu_flist_alloc_root(struct ukfs *fs, const char *path)
{
	FSU_FENT *root;
	char *p;
	int rv;

	root = malloc(sizeof(FSU_FENT));
	if (root == NULL) {
		warn("malloc");
		return NULL;
	}

	root->parent = root->next = NULL;
	root->childno = 0;
	root->path = strdup(path);
	root->size = 0;
	root->done = false;
	root->level = 0;

	p = strrchr(root->path, '/');
	if (p == NULL || (p == root->path && root->path[1] == '\0'))
		root->filename = root->path;
	else
		root->filename = p + 1;

	root->pathlen = strlen(root->path);
	root->isdot = false;

	rv = statfun(fs, root->path, &root->sb);
	if (rv == -1) {
		warn("%s", root->path);
		fsu_flist_free_entry(root);
		return NULL;
	}
	return root;
}

void
fsu_flist_free(FSU_FENT *root)
{
	FSU_FENT *cur;

	while ((cur = root) != NULL) {
		root = root->next;
		free(cur->path);
		free(cur);		
	}
}

void
fsu_flist_free_entry(FSU_FENT *ent)
{

	if (ent == NULL)
		return;

	ent->next = NULL;
	free(ent->path);
	free(ent);
}

static void
fsu_flist_sort(FSU_FENT **firstp, FSU_FENT **lastp, int nbelt)
{
	FSU_FENT *p;
	int i;

	if (nbelt > array_size) {
		array = realloc(array, (nbelt + 20) * sizeof(FSU_FENT *));
		if (array == NULL)
			return;
		array_size = nbelt + 20;
	}

	for (i=0, p = *firstp; p != NULL; p=p->next, i++)
		array[i] = p;
	
	qsort((void *)array, nbelt, sizeof(FSU_FENT *), cmp);
	
	*firstp = *lastp = array[0];
	for (i = 1; i < nbelt; ++i)
		*lastp = (*lastp)->next = array[i];
	(*lastp)->next = NULL;
}
