/* $Id: fsu_mount.c,v 1.1 2008/07/09 10:35:30 stacktic Exp $ */

/*
 * Copyright (c) 2008 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <string.h>

#include <sys/mount.h>

#include <isofs/cd9660/cd9660_mount.h>
#include <fs/efs/efs_mount.h>
#include <ufs/ufs/ufsmount.h> /* ext2fs, ffs, lfs */
#include <fs/hfs/hfs.h>
#include <fs/msdosfs/msdosfsmount.h>
#include <ntfs/ntfsmount.h>
#include <fs/tmpfs/tmpfs.h>
#include <fs/udf/udf_mount.h>

#include "ukfs.h"

#include "fsu_mount.h"

union fsu_args {
	struct iso_args args_cd9660;
	struct efs_args args_efs;
	struct ufs_args args_ufs; /* ext2fs, ffs, lfs */
	struct hfs_args args_hfs;
	struct msdosfs_args args_msdos;
	struct ntfs_args args_ntfs;
	struct tmpfs_args args_tmpfs;
	struct udf_args args_udf;
};

struct ukfs *
fsu_mount(const char *fstype, const char *fsdevice, int mntflags)
{
	union fsu_args args;
	void *argsp = NULL;
	int argssize = 0;
	struct ukfs *fs;

	memset(&args, 0, sizeof(args));

	if (strcmp(fstype, "tmpfs") == 0) {
		ukfs_init();
		args.args_tmpfs.ta_version = TMPFS_ARGS_VERSION;
		args.args_tmpfs.ta_root_mode = 0777;
		return ukfs_mount(fstype, fsdevice, "/", mntflags, 
				  &args.args_tmpfs, 
				  sizeof(struct tmpfs_args));
	}

	args.args_ufs.fspec = strdup(fsdevice);
	
	if (strcmp(fstype, "cd9660") == 0) {
		if ((mntflags & MNT_RDONLY) == 0) {
			fprintf(stderr, "cd9660 file-systems can only be"
				" mounted with read-only flag\n");
			fprintf(stderr, "mounting read-only\n");
			mntflags |= MNT_RDONLY;
		}
		argsp = (struct iso_args*)&args.args_cd9660;
		argssize = sizeof(struct iso_args);
	} else if (strcmp(fstype, "efs") == 0) {
		argsp = (struct efs_args*)&args.args_efs;
		argssize = sizeof(struct efs_args);
	} else if (strcmp(fstype, "ext2fs") == 0 || 
		   strcmp(fstype, "ffs") == 0 ||
		   strcmp(fstype, "lfs") == 0) {
		argsp = (struct ufs_args*)&args.args_ufs;
		argssize = sizeof(struct ufs_args);
	} else if (strcmp(fstype, "hfs") == 0) {
		argsp = (struct hfs_args*)&args.args_hfs;
		argssize = sizeof(struct hfs_args);
	} else if (strcmp(fstype, "msdos") == 0) {
		argsp = (struct msdosfs_args*)&args.args_msdos;
		argssize = sizeof(struct msdosfs_args);
	} else if (strcmp(fstype, "ntfs") == 0) {
		argsp = (struct ntfs_args*)&args.args_ntfs;
		argssize = sizeof(struct ntfs_args);
	} else if (strcmp(fstype, "udf") == 0) {
		argsp = (struct udf_args*)&args.args_udf;
		argssize = sizeof(struct udf_args);
	} else
		return NULL;

	ukfs_init();
	fs = ukfs_mount(fstype, fsdevice, "/", mntflags, argsp, argssize);
	free(args.args_ufs.fspec);
	return fs;
}
